// packages/@code-like-a-carpenter/foundation-parser/src/extractors/cdc.ts
import assert from "assert";
import crypto from "crypto";
import path from "path";
import { assertObjectType } from "graphql";
import camelCase from "lodash/camelCase.js";
import kebabCase from "lodash/kebabCase.js";
import snakeCase from "lodash/snakeCase.js";
import upperFirst from "lodash/upperFirst.js";
import { DispatcherConfigSchema, HandlerConfigSchema } from "../config.mjs";
import {
  filterNull,
  getArgEnumValue,
  getArgStringValue,
  getOptionalArg,
  getOptionalArgBooleanValue,
  getOptionalArgObjectValue
} from "../helpers.mjs";
import {
  resolveActionsModuleId,
  resolveDependenciesModuleId,
  resolveHandlerModuleId
} from "../paths.mjs";
import { extractTableName } from "../tables.mjs";
function extractDispatcherConfig(config, outputFile, type) {
  const tableName = extractTableName(type);
  const filename = `dispatcher-${kebabCase(tableName)}`;
  const functionName = `${tableName}CDCDispatcher`;
  const directory = path.join(path.dirname(outputFile), filename);
  assert(
    type?.astNode?.directives,
    `Expected to find at least one directive on type ${type.name}`
  );
  const { batchSize, maximumRetryAttempts, memorySize, timeout } = type.astNode.directives.filter((d) => d.name.value === "reacts" || d.name.value === "enriches").map(
    (directive) => DispatcherConfigSchema.parse({
      ...config.dispatcherDefaults,
      ...getOptionalArgObjectValue("dispatcherConfig", directive)
    })
  ).reduce((acc, next) => {
    return {
      batchSize: Math.max(acc.batchSize, next.batchSize),
      maximumRetryAttempts: Math.max(
        acc.maximumRetryAttempts,
        next.maximumRetryAttempts
      ),
      memorySize: Math.max(acc.memorySize, next.memorySize),
      timeout: Math.max(acc.timeout, next.timeout)
    };
  }, config.dispatcherDefaults);
  const nestedStackLocation = "./dispatcher.yml";
  return {
    batchSize,
    dependenciesModuleId: resolveDependenciesModuleId(config, directory),
    directory,
    filename,
    functionName,
    maximumRetryAttempts,
    memorySize,
    nestedStackLocation,
    runtimeModuleId: "@code-like-a-carpenter/foundation-runtime",
    timeout
  };
}
function extractChangeDataCaptureConfig(config, schema, type, outputFile) {
  return type.astNode?.directives?.map((directive) => {
    if (directive.name.value === "enriches") {
      return extractEnricherConfig(
        config,
        schema,
        type,
        directive,
        outputFile
      );
    }
    if (directive.name.value === "reacts") {
      return extractReactorConfig(
        config,
        schema,
        type,
        directive,
        outputFile
      );
    }
    if (directive.name.value === "reduces") {
      return extractReducerConfig(
        config,
        schema,
        type,
        directive,
        outputFile
      );
    }
    return null;
  }).filter(filterNull) ?? [];
}
function getTargetTable(schema, modelName, produces) {
  const targetModel = schema.getType(produces);
  assert(
    targetModel,
    `\`targetModel\` arg on CDC directive for ${modelName} identifies ${produces}, which does not appear to identify a type`
  );
  return extractTableName(assertObjectType(targetModel));
}
function getTargetTables(fieldName, schema, directive) {
  const arg = getOptionalArg(fieldName, directive);
  if (!arg) {
    return [];
  }
  assert(arg.value.kind === "ListValue", `Expected ${fieldName} to be a list`);
  return Array.from(
    new Set(
      arg.value.values.map((v) => {
        assert(
          v.kind === "StringValue",
          `Expected @${directive.name.value} directive argument "${fieldName}" to be a list of strings`
        );
        return getTargetTable(schema, directive.name.value, v.value);
      })
    )
  );
}
function extractCommonConfig(config, schema, type, directive, outputFile, filename) {
  const sourceModelName = type.name;
  const handlerImportName = getArgStringValue("importName", directive);
  const handler = getArgStringValue("importPath", directive);
  const directory = path.join(path.dirname(outputFile), filename);
  const { memorySize, timeout } = HandlerConfigSchema.parse({
    ...config.handlerDefaults,
    ...getOptionalArgObjectValue("handlerConfig", directive)
  });
  const nestedStackLocation = "./cdc.yml";
  return {
    actionsModuleId: resolveActionsModuleId(config, directory),
    directory,
    handlerImportName,
    handlerModuleId: resolveHandlerModuleId(type, directory, handler),
    memorySize,
    nestedStackLocation,
    runtimeModuleId: "@code-like-a-carpenter/foundation-runtime",
    sourceModelName,
    timeout
  };
}
function extractEnricherConfig(config, schema, type, directive, outputFile) {
  const event = getEvent(type, directive);
  const sourceModelName = type.name;
  const targetModelName = getArgStringValue("targetModel", directive);
  const filename = `enricher--${kebabCase(
    sourceModelName
  )}--${event.toLowerCase()}--${kebabCase(targetModelName)}`;
  const functionName = makeFunctionName(
    "enricher",
    sourceModelName,
    event,
    targetModelName
  );
  const readableTables = getTargetTables("readableModels", schema, directive);
  const writableTables = Array.from(
    /* @__PURE__ */ new Set([
      ...getTargetTables("writableModels", schema, directive),
      getTargetTable(schema, type.name, targetModelName)
    ])
  );
  return {
    ...extractCommonConfig(
      config,
      schema,
      type,
      directive,
      outputFile,
      filename
    ),
    event,
    filename,
    functionName,
    readableTables,
    targetModelName,
    type: "ENRICHER",
    writableTables
  };
}
function extractReactorConfig(config, schema, type, directive, outputFile) {
  const event = getEvent(type, directive);
  const readableTables = getTargetTables("readableModels", schema, directive);
  const writableTables = getTargetTables("writableModels", schema, directive);
  const sourceModelName = type.name;
  const filename = `react--${kebabCase(
    sourceModelName
  )}--${event.toLowerCase()}`;
  const functionName = makeFunctionName("react", sourceModelName, event);
  return {
    ...extractCommonConfig(
      config,
      schema,
      type,
      directive,
      outputFile,
      filename
    ),
    event,
    filename,
    functionName,
    readableTables,
    sourceModelName: type.name,
    type: "REACTOR",
    writableTables
  };
}
function extractReducerConfig(config, schema, type, directive, outputFile) {
  const event = getEvent(type, directive);
  const sourceModelName = type.name;
  const targetModelName = getArgStringValue("targetModel", directive);
  const filename = `reduce--${kebabCase(
    sourceModelName
  )}--${event.toLowerCase()}--${kebabCase(targetModelName)}`;
  const functionName = makeFunctionName(
    "reduce",
    sourceModelName,
    event,
    targetModelName
  );
  const readableTables = Array.from(
    /* @__PURE__ */ new Set([
      ...getTargetTables("readableModels", schema, directive),
      extractTableName(assertObjectType(type))
    ])
  );
  const writableTables = Array.from(
    /* @__PURE__ */ new Set([
      ...getTargetTables("writableModels", schema, directive),
      getTargetTable(schema, type.name, targetModelName)
    ])
  );
  return {
    ...extractCommonConfig(
      config,
      schema,
      type,
      directive,
      outputFile,
      filename
    ),
    event,
    filename,
    functionName,
    multiReduce: getOptionalArgBooleanValue("multiReduce", directive) ?? false,
    readableTables,
    sourceModelName: type.name,
    targetModelName,
    type: "REDUCER",
    writableTables
  };
}
function getEvent(type, directive) {
  const event = getArgEnumValue("event", directive);
  assert(
    event === "INSERT" || event === "MODIFY" || event === "REMOVE" || event === "UPSERT",
    `Invalid event type ${event} for @${directive.name.value} on ${type.name}`
  );
  return event;
}
function makeFunctionName(prefix, sourceModelName, event, targetModelName = "") {
  const sourceAbbr = abbreviate(sourceModelName);
  const targetAbbr = abbreviate(targetModelName);
  const rand = crypto.createHash("sha1").update(sourceModelName + event + targetModelName).digest("hex").slice(0, 8);
  const functionName = `Fn${upperFirst(
    camelCase(
      [prefix, sourceAbbr, event, targetAbbr].filter(filterNull).join("--")
    )
  )}${rand}`;
  assert(
    functionName.length <= 64,
    `Handler function name must be less than 64 characters: ${functionName}`
  );
  return functionName;
}
function abbreviate(str) {
  return snakeCase(str).split("_").map((c) => c[0]).join("-");
}
export {
  extractChangeDataCaptureConfig,
  extractDispatcherConfig,
  getTargetTables
};
//# sourceMappingURL=cdc.mjs.map
