// packages/@code-like-a-carpenter/foundation-parser/src/tables.ts
import { assert } from "@code-like-a-carpenter/assert";
import { extractDispatcherConfig } from "./extractors/cdc.mjs";
import {
  getOptionalArgBooleanValue,
  getOptionalArgStringValue,
  getOptionalDirective,
  hasInterface
} from "./helpers.mjs";
import { getModel } from "./models.mjs";
import { resolveDependenciesModuleId } from "./paths.mjs";
var schemaTables = /* @__PURE__ */ new WeakMap();
function extractTable(config, schema, type, outputFile) {
  let tables = schemaTables.get(schema);
  if (!tables) {
    tables = /* @__PURE__ */ new Map();
    schemaTables.set(schema, tables);
  }
  const tableName = extractTableName(type);
  const existingTable = tables.get(tableName);
  if (existingTable) {
    return updateExistingTable(
      config,
      type,
      existingTable,
      tableName,
      outputFile
    );
  }
  const newTable = defineNewTable(config, type, tableName, outputFile);
  tables.set(newTable.tableName, newTable);
  return newTable;
}
function extractTableName(type) {
  const directive = getOptionalDirective("table", type);
  if (directive) {
    const value = getOptionalArgStringValue("name", directive);
    if (value) {
      return value;
    }
  }
  return `Table${type.name}`;
}
function assertPrimaryKeysMatch(tableName, tablePrimaryKey, modelPrimaryKey) {
  assert(
    tablePrimaryKey.isComposite === modelPrimaryKey.isComposite,
    `All models in table must be composite (or not). Please check the definitions for the models in ${tableName}.`
  );
}
function combineSecondaryIndexes(tableName, tableIndexes, modelIndexes) {
  const tableIndexMap = new Map(
    tableIndexes.map((t) => [t.name, t])
  );
  const modelIndexMap = new Map(
    modelIndexes.map((m) => [m.name, m])
  );
  for (const [name, modelIndex] of modelIndexMap.entries()) {
    const index = tableIndexMap.get(name);
    if (index) {
      if (index.type === "gsi" && modelIndex.type === "gsi") {
        assert(
          index.isComposite === modelIndex.isComposite,
          `Please check the secondary index ${name} for the table ${tableName}. All indexes of the same name must be of the same type (either partition or composite).`
        );
      }
      assert(
        index.type === modelIndex.type,
        `Please check the secondary index ${name} for the table ${tableName}. All indexes of the same name must be of the same type (either gsi or lsi).`
      );
      assert(
        index.projectionType === modelIndex.projectionType,
        `Please check the secondary index ${name} for the table ${tableName}. All indexes of the same name must be of the same projection type (either "all"" or "keys_only").`
      );
    } else {
      tableIndexes.push(modelIndexToTableIndex(modelIndex));
    }
  }
}
function defineNewTable(config, type, tableName, outputFile) {
  const model = getModel(type);
  const dependenciesModuleId = resolveDependenciesModuleId(config, outputFile);
  const table = {
    dependenciesModuleId,
    enablePointInTimeRecovery: shouldEnablePointInTimeRecovery(config, type),
    enableStreaming: shouldEnableStreaming(type, model),
    hasPublicModels: hasInterface("PublicModel", type),
    hasTtl: !!model.ttlConfig,
    primaryKey: model.primaryKey,
    secondaryIndexes: model.secondaryIndexes.map(modelIndexToTableIndex),
    tableName
  };
  if (model.changeDataCaptureConfig.length) {
    return {
      ...table,
      dispatcherConfig: extractDispatcherConfig(config, outputFile, type),
      hasCdc: true
    };
  }
  return { ...table, hasCdc: false };
}
function modelIndexToTableIndex(mi) {
  if (mi.type === "lsi") {
    return {
      isComposite: mi.isComposite,
      name: mi.name,
      partitionKeyIsSingleField: mi.partitionKeyIsSingleField,
      partitionKeyName: mi.partitionKeyName,
      projectionType: mi.projectionType,
      sortKeyIsSingleField: mi.sortKeyIsSingleField,
      sortKeyName: mi.sortKeyName,
      type: "lsi"
    };
  }
  if (mi.isComposite) {
    return {
      isComposite: mi.isComposite,
      name: mi.name,
      partitionKeyIsSingleField: mi.partitionKeyFields.length === 1,
      partitionKeyName: mi.partitionKeyName,
      projectionType: mi.projectionType,
      sortKeyIsSingleField: mi.isComposite && mi.sortKeyFields.length === 1,
      sortKeyName: mi.sortKeyName,
      type: "gsi"
    };
  }
  return {
    isComposite: mi.isComposite,
    name: mi.name,
    partitionKeyIsSingleField: mi.partitionKeyFields.length === 1,
    partitionKeyName: mi.partitionKeyName,
    partitionKeyPrefix: mi.partitionKeyPrefix,
    projectionType: mi.projectionType,
    type: "gsi"
  };
}
function updateExistingTable(config, type, existingTable, tableName, outputFile) {
  const model = getModel(type);
  existingTable.enablePointInTimeRecovery = existingTable.enablePointInTimeRecovery || shouldEnablePointInTimeRecovery(config, type);
  existingTable.enableStreaming = existingTable.enableStreaming || shouldEnableStreaming(type, model);
  existingTable.hasPublicModels = existingTable.hasPublicModels || hasInterface("PublicModel", type);
  if (existingTable.hasCdc || model.changeDataCaptureConfig.length > 0) {
    existingTable.hasCdc = true;
    existingTable.dispatcherConfig = extractDispatcherConfig(
      config,
      outputFile,
      type
    );
  }
  existingTable.hasTtl = existingTable.hasTtl || !!model.ttlConfig;
  assertPrimaryKeysMatch(tableName, existingTable.primaryKey, model.primaryKey);
  combineSecondaryIndexes(
    tableName,
    existingTable.secondaryIndexes,
    model.secondaryIndexes
  );
  return existingTable;
}
function shouldEnablePointInTimeRecovery(config, type) {
  const tableDirective = getOptionalDirective("table", type);
  if (!tableDirective) {
    return config.tableDefaults.enablePointInTimeRecovery;
  }
  return getOptionalArgBooleanValue("enablePointInTimeRecovery", tableDirective) !== false;
}
function shouldEnableStreaming(type, model) {
  if (model.changeDataCaptureConfig.length > 0) {
    return true;
  }
  const tableDirective = getOptionalDirective("table", type);
  if (!tableDirective) {
    return false;
  }
  return getOptionalArgBooleanValue("enableStreaming", tableDirective) === true;
}
export {
  assertPrimaryKeysMatch,
  extractTable,
  extractTableName
};
//# sourceMappingURL=tables.mjs.map
